﻿const co = require('co');

module.exports = createExpressAsync;

const expressMethods =
[
  'all', 'checkout', 'copy', 'delete', 'get',
  'head', 'lock', 'merge', 'mkactivity', 'mkcol',
  'move', 'm-search', 'notify', 'options', 'patch',
  'post', 'purge', 'put', 'report', 'search',
  'subscribe', 'trace', 'unlock', 'unsubscribe', 'use'
];

function handleAsync(handler) {
  return (req, res, next) => co(function*() {
      yield handler(req, res, next);
    })
    .catch(next);
}

function createAsyncMethod(express, method) {
  return function doAsync(...args) {
    const asyncHandler = args.pop();
    const expressHandler = handleAsync(asyncHandler);
    args.push(expressHandler);
    express[method](...args);
  };
}

function createDescriptor(value) {
  return {
    configurable: false,
    enumerable: true,
    value,
    writable: false
  };
}

function patchExpressAsync(express) {
  for (const method of expressMethods.filter(m => m in express)) {
    express[`${method}Async`] = createAsyncMethod(express, method);
  }
  return express;
}

function createExpressAsync(express) {
  const descriptorMap = {};

  for (const method of expressMethods.filter(m => m in express)) {
    descriptorMap[method] = createDescriptor(createAsyncMethod(express, method));
  }

  descriptorMap.express = createDescriptor(express);
  const expressAsync = Object.create({}, descriptorMap);
  return expressAsync;
};

/*function createExpressAsync(express) {
  const descriptorMap = {};

  for (const method of expressMethods.filter(m => m in express)) {
    descriptorMap[`${method}Async`] = createDescriptor(createAsyncMethod(express, method));
  }
  if (typeof express.listen === "function") descriptorMap.listen = createDescriptor(express.listen.bind(express));
  descriptorMap.express = createDescriptor(express);

  const expressAsync = Object.create(express, descriptorMap);
  return expressAsync;
};*/